from ._async_utils import EventLoopGetter
from ._loader import get_celery_loader
from ._provide import PROVIDE
from ._task import BaseCeleryConnectableDeps, InjectableCeleryTask

__all__ = (
    "InjectableCeleryTask",
    "BaseCeleryConnectableDeps",
    "get_celery_loader",
    "EventLoopGetter",
    "PROVIDE",
)



from __future__ import annotations

import functools
import inspect
from dataclasses import dataclass
from threading import Lock
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

if TYPE_CHECKING:
    from collections.abc import Iterable

    from magic_di import ConnectableProtocol

T = TypeVar("T")


@dataclass(frozen=True)
class Dependency(Generic[T]):
    object: type[T]
    instance: T | None = None


class SingletonDependencyContainer:
    def __init__(self) -> None:
        self._deps: dict[type, Dependency[Any]] = {}
        self._lock: Lock = Lock()

    def add(self, obj: type[T], **kwargs: Any) -> type[T]:
        with self._lock:
            if dep := self._get(obj):
                return dep

        wrapped = _wrap(obj, **kwargs)
        dependency = Dependency(
            object=wrapped,
            instance=wrapped() if inspect.isclass(obj) else None,
        )

        with self._lock:
            # To avoid deadlocks when an injector is used inside the __init__ method
            if dep := self._get(obj):
                return dep

            self._deps[obj] = dependency

        return wrapped

    def get(self, obj: type[T]) -> type[T] | None:
        with self._lock:
            return self._get(obj)

    def iter_instances(
        self,
        *,
        reverse: bool = False,
    ) -> Iterable[tuple[type, ConnectableProtocol]]:
        with self._lock:
            deps_iter: Iterable[Dependency[Any]] = list(
                reversed(self._deps.values()) if reverse else self._deps.values(),
            )

        for dep in deps_iter:
            if dep.instance:
                yield dep.object, dep.instance

    def _get(self, obj: type[T]) -> type[T] | None:
        dep = self._deps.get(obj)
        return dep.object if dep else None


def _wrap(obj: type[T], *args: Any, **kwargs: Any) -> type[T]:
    if not inspect.isclass(obj):
        partial = functools.wraps(obj)(functools.partial(obj, *args, **kwargs))
        return cast(type[T], partial)

    _instance: T | None = None

    def new(_: Any) -> T:
        nonlocal _instance

        if _instance is not None:
            return _instance

        _instance = obj(*args, **kwargs)
        return _instance

    # This class wraps obj and creates a singleton class
    # that uses *args and **kwargs for initialization.
    #
    # Please note that it also doesn't allow passing additional args
    # after creating this partial.
    # It was made to make it more obvious that you can't create
    # two different instances with different parameters.
    #
    # Example:
    #   >>> injected_redis = injector.inject(Redis)
    #   >>> injected_redis()  # works
    #   >>> injected_redis(timeout=10)  # doesn't work
    #
    # Here we manually create a new singleton class factory using the `type` metaclass
    # Since the original class was not modified, it will use its own metaclass.
    return functools.wraps(
        obj,
        updated=(),
    )(
        type(
            obj.__name__,
            (),
            {"__new__": new},
        ),
    )



"""
Package containing tools for integrating Dependency Injector with FastAPI framework.
"""

from ._app import inject_app
from ._provide import Provide, Provider  # type: ignore[attr-defined]

__all__ = ("inject_app", "Provide", "Provider")



from __future__ import annotations

from typing import Any, TypeVar, cast, get_args
from typing import get_type_hints as _get_type_hints

from magic_di import ConnectableProtocol

LegacyUnionType = type(object | None)

try:
    from types import UnionType  # type: ignore[import-error,unused-ignore]
except ImportError:
    UnionType = LegacyUnionType  # type: ignore[misc]


T = TypeVar("T")

NONE_TYPE: type = type(None)


def is_class(obj: Any) -> bool:
    return isinstance(obj, type)


def get_cls_from_optional(cls: T) -> T:
    """
    Extract the actual class from a union that includes None.
    If it is not a union type hint, it returns the same type hint.
    Example:
        >>> get_cls_from_optional(Union[str, None])
        str
        >>> get_cls_from_optional(str | None)
        str
        >>> get_cls_from_optional(str | None)
        str
        >>> get_cls_from_optional(int)
        int
    Args:
        cls (T): Type hint for class
    Returns:
        T: Extracted class
    """
    if not isinstance(cls, UnionType | LegacyUnionType):
        return cls

    args = get_args(cls)

    optional_type_hint_args_len = 2
    if len(args) != optional_type_hint_args_len:
        return cast(T, cls)

    if NONE_TYPE not in args:
        return cast(T, cls)

    for typo in args:
        if not safe_is_subclass(typo, NONE_TYPE):
            return cast(T, typo)

    return cast(T, cls)


def safe_is_subclass(sub_cls: Any, cls: type) -> bool:
    try:
        return issubclass(sub_cls, cls)
    except TypeError:
        return False


def safe_is_instance(sub_cls: Any, cls: type) -> bool:
    try:
        return isinstance(sub_cls, cls)
    except TypeError:
        return False


def is_injectable(cls: Any) -> bool:
    """
    Check if a class is a subclass of ConnectableProtocol.

    Args:
        cls (Any): The class to check.

    Returns:
        bool: True if the class is a subclass of ConnectableProtocol, False otherwise.
    """
    return safe_is_subclass(cls, ConnectableProtocol) or safe_is_instance(cls, ConnectableProtocol)


def get_type_hints(obj: Any, *, include_extras: bool = False) -> dict[str, type]:
    try:
        if is_class(obj):
            return _get_type_hints(obj.__init__, include_extras=include_extras)

        return _get_type_hints(obj, include_extras=include_extras)
    except TypeError:
        return {}



from __future__ import annotations

import inspect
from functools import wraps
from typing import TYPE_CHECKING, Any, cast, get_type_hints

from celery.app.task import Task

from magic_di import Connectable, DependencyInjector
from magic_di.celery._async_utils import EventLoop, run_in_event_loop
from magic_di.celery._loader import InjectedCeleryLoaderProtocol

if TYPE_CHECKING:
    from collections.abc import Callable


class BaseCeleryConnectableDeps(Connectable): ...


class InjectableCeleryTaskMetaclass(type):
    def __new__(cls, name: str, bases: tuple[type, ...], dct: dict[str, Any]) -> type:
        run = dct.get("run")
        run_wrapper = dct.get("run_wrapper")

        if not run:
            return super().__new__(cls, name, bases, dct)

        if run_wrapper:
            dct["run"] = run_wrapper(run)
        else:
            for base in bases:
                if hasattr(base, "run_wrapper"):
                    dct["run"] = base.run_wrapper(run)
                    break

        return super().__new__(cls, name, bases, dct)


class InjectableCeleryTask(Task, Connectable, metaclass=InjectableCeleryTaskMetaclass):  # type: ignore[type-arg]
    __annotations__ = {}

    def __init__(
        self,
        # Type hint Any to make it non-injectable
        deps_instance: Any = None,
        injector: DependencyInjector | None = None,
    ) -> None:
        self._injector = injector

        if deps_instance:
            self.deps = deps_instance
            return

        if not self.injector:
            error_msg = (
                "Celery is not injected by magic_di.celery.get_celery_loader "
                "and injector or deps_instance is not passed in task args"
            )
            raise ValueError(error_msg)

        for dep in self.injector.inspect(self.run).deps.values():
            self.injector.lazy_inject(dep)

        deps_cls = get_type_hints(self).get("deps", BaseCeleryConnectableDeps)
        self.deps = self.injector.inject(deps_cls)()

        super().__init__()

    @property
    def injector(self) -> DependencyInjector | None:
        if self._injector:
            return self._injector

        if isinstance(self.app.loader, InjectedCeleryLoaderProtocol):
            loader: InjectedCeleryLoaderProtocol = cast(
                InjectedCeleryLoaderProtocol,
                self.app.loader,
            )

            return loader.injector

        return None

    def load(self) -> None:
        if isinstance(self.app.loader, InjectedCeleryLoaderProtocol):
            loader: InjectedCeleryLoaderProtocol = cast(
                InjectedCeleryLoaderProtocol,
                self.app.loader,
            )
            loader.on_worker_process_init()
            return

        self.app.loader.on_worker_process_init()  # type: ignore[attr-defined]

    @property
    def loaded(self) -> bool:
        if not isinstance(self.app.loader, InjectedCeleryLoaderProtocol):
            return True

        loader: InjectedCeleryLoaderProtocol = cast(
            InjectedCeleryLoaderProtocol,
            self.app.loader,
        )
        return loader.loaded

    def get_event_loop(self) -> EventLoop | None:
        if isinstance(self.app.loader, InjectedCeleryLoaderProtocol):
            loader: InjectedCeleryLoaderProtocol = cast(
                InjectedCeleryLoaderProtocol,
                self.app.loader,
            )

            return loader.get_event_loop()

        return None

    @staticmethod
    def run_wrapper(orig_run: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(orig_run)
        def runner(self: InjectableCeleryTask, *args: Any, **kwargs: Any) -> Any:
            if not self.loaded:
                self.load()

            run = self.injector.inject(orig_run) if self.injector else orig_run

            if "self" in inspect.signature(run).parameters:
                args = (self, *args)

            result = run(*args, **kwargs)

            if inspect.isawaitable(result):
                event_loop = self.get_event_loop()
                if not event_loop:
                    error_msg = (
                        "Cannot run async tasks. "
                        "Celery is not injected by magic_di.celery.get_celery_loader"
                    )
                    raise ValueError(error_msg)

                return run_in_event_loop(
                    result,  # type: ignore[arg-type]
                    event_loop,
                )

            return result

        return runner



from __future__ import annotations

import inspect
import logging
from contextlib import contextmanager
from threading import Lock
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    TypeVar,
    cast,
    get_origin,
)

from magic_di._container import SingletonDependencyContainer
from magic_di._signature import Signature
from magic_di._utils import (
    get_cls_from_optional,
    get_type_hints,
    is_injectable,
    safe_is_subclass,
)
from magic_di.exceptions import InjectionError, InspectionError

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator

    from magic_di._connectable import ConnectableProtocol

# flag to use in typing.Annotated
# to forcefully mark dependency as injectable
# even if dependency is not connectable
# Example:
# >>> typing.Annotated[MyDependency, Injectable]
Injectable = type("Injectable", (), {})()


def is_injector(cls: type) -> bool:
    """
    Check if a class is a subclass of DependencyInjector.

    Args:
        cls (type): The class to check.

    Returns:
        bool: True if the class is a subclass of DependencyInjector, False otherwise.
    """
    return safe_is_subclass(cls, DependencyInjector)


def is_forcefully_marked_as_injectable(cls: Any) -> bool:
    if get_origin(cls) is Annotated:
        return Injectable in cls.__metadata__

    return False


T = TypeVar("T")
AnyObject = TypeVar("AnyObject")

logger = logging.getLogger("injector")


class DependencyInjector:
    def __init__(
        self,
        bindings: dict[type, type] | None = None,
        logger: logging.Logger = logger,
    ):
        self.bindings = bindings or {}
        self.logger: logging.Logger = logger

        self._deps = SingletonDependencyContainer()
        self._postponed: list[Callable[..., T]] = []
        self._lock = Lock()

    def inject(self, obj: Callable[..., T]) -> Callable[..., T]:
        """
        Inject dependencies into a class/function.
        This method is idempotent, always returns the same instance

        Args:
            obj (Callable[..., T]): The class/function to inject dependencies into.

        Returns:
            Callable[..., T]: Partial for class/function with dependencies injected.
        """
        obj = self._unwrap_type_hint(obj)  # type: ignore[arg-type]

        if dep := self._deps.get(obj):
            return dep

        signature = self.inspect(obj)

        clients: dict[str, object] = {}
        for name, dep in signature.deps.items():
            clients[name] = self.inject(dep)()

        if signature.injector_arg is not None:
            clients[signature.injector_arg] = self

        try:
            return self._deps.add(obj, **clients)
        except TypeError as exc:
            raise InjectionError(obj, signature) from exc

    def inspect(self, obj: AnyObject) -> Signature[AnyObject]:
        try:
            hints: dict[str, type[Any]] = get_type_hints(obj)
            hints_with_extras = get_type_hints(obj, include_extras=True)

            if not hints:
                return Signature(obj, is_injectable=is_injectable(obj))

            if inspect.ismethod(obj):
                hints.pop("self", None)

            hints.pop("return", None)

            signature = Signature(obj, is_injectable=is_injectable(obj))

            for name, hint_ in hints.items():
                hint = self._unwrap_type_hint(hint_)
                hint_with_extra = hints_with_extras[name]

                if is_injector(hint):
                    signature.injector_arg = name
                elif not is_injectable(hint) and not is_forcefully_marked_as_injectable(
                    hint_with_extra,
                ):
                    signature.kwargs[name] = hint
                else:
                    signature.deps[name] = hint

        except Exception as exc:  # noqa: BLE001
            raise InspectionError(obj) from exc

        return signature

    async def connect(self) -> None:
        """
        Connect all injected dependencies
        """
        # unpack to create copy of list
        for postponed in [*self._postponed]:
            # First, we need to create instances of postponed injection
            # in order to connect them.
            self.inject(postponed)

        for cls, instance in self._deps.iter_instances():
            if is_injectable(instance):
                self.logger.debug("Connecting %s...", cls.__name__)
                await instance.__connect__()

    async def disconnect(self) -> None:
        """
        Disconnect all injected dependencies
        """
        for cls, instance in self._deps.iter_instances(reverse=True):
            if is_injectable(instance):
                try:
                    await instance.__disconnect__()
                except Exception:
                    self.logger.exception("Failed to disconnect %s", cls.__name__)

    async def __aenter__(self) -> DependencyInjector:  # noqa: PYI034
        await self.connect()
        return self

    async def __aexit__(self, *args: object, **kwargs: Any) -> None:
        await self.disconnect()

    def iter_deps(self) -> Iterable[ConnectableProtocol]:
        instance: ConnectableProtocol

        for _, instance in self._deps.iter_instances():
            yield instance

    def lazy_inject(self, obj: Callable[..., T]) -> Callable[..., T]:
        """
        Lazily inject dependencies into a class or function.

        Args:
            obj (Callable[..., T]): The class or function to inject dependencies into.

        Returns:
            Callable[..., T]: A function that, when called,
            will inject the dependencies and
            return the injected class instance or function.
        """

        with self._lock:
            self._postponed.append(obj)  # type: ignore[arg-type]
            # incompatible type "Callable[..., T]"; expected "Callable[..., T]"

        injected: T | None = None

        def inject() -> T:
            nonlocal injected

            if injected is not None:
                return injected

            injected = self.inject(obj)()
            return injected

        return cast(type[T], inject)

    def bind(self, bindings: dict[type, type]) -> None:
        """
        Bind new bindings to the injector.

        This method is used to add new bindings to the injector.
        Bindings are a dictionary where the keys are the
        classes used in dependencies type hints,
        and the values are the classes that should replace them.

        For example, if you have a class `Foo` that depends on an interface `Bar`,
        and you have a class `BarImpl` that implements `Bar`,
        you would add a binding like this: `injector.bind({Bar: BarImpl})`.
        Then, whenever `Foo` is injected,
        it will receive an instance of `BarImpl` instead of `Bar`.

        If a binding for a particular class or type already exists,
        this method will update that binding with the new value

        Args:
            bindings (dict[type, type]): The bindings to add.
            This should be a dictionary where the keys are
            classes and the values are the classes that should replace them.
        """
        with self._lock:
            self.bindings = self.bindings | bindings

    @contextmanager
    def override(self, bindings: dict[type, type]) -> Iterator[None]:
        """
        Temporarily override the bindings and dependencies of the injector.

        Args:
            bindings (dict): The bindings to use for the override.
        """
        with self._lock:
            actual_deps = self._deps
            actual_bindings = self.bindings

            self._deps = SingletonDependencyContainer()
            self.bindings = self.bindings | bindings

        try:
            yield
        finally:
            with self._lock:
                self._deps = actual_deps
                self.bindings = actual_bindings

    def _unwrap_type_hint(self, obj: type[AnyObject]) -> type[AnyObject]:
        obj = get_cls_from_optional(obj)
        return self.bindings.get(obj, obj)

    def __hash__(self) -> int:
        """Injector is always unique"""
        return id(self)



# An empty object that returns an ellipsis
# It's used only to pass celery's strict_typing check.
# It is not necessary to use Provide()
# if you disable the `strict_typing` option in Celery's configuration.
PROVIDE = type("DependencyProvider", (), {})()



import inspect
from collections.abc import Callable
from typing import Any

from magic_di._signature import Signature
from magic_di._utils import get_type_hints


class InjectorError(Exception): ...


class InjectionError(InjectorError):
    def __init__(self, obj: Callable[..., Any], signature: Signature[Any]):
        self.obj = obj
        self.signature = signature

    def __str__(self) -> str:
        return self._build_error_message()

    def _build_error_message(self) -> str:
        try:
            full_signature = inspect.signature(self.obj)
            type_hints = get_type_hints(self.obj)

            missing_kwargs = {
                param.name: type_hints.get(param.annotation) or param.annotation
                for param in full_signature.parameters.values()
                if param.name not in self.signature.deps and param.default is full_signature.empty
            }

            missing_kwargs_str = "\n".join(
                [f"{name}: {hint!r}" for name, hint in missing_kwargs.items()],
            )

            object_location = _get_object_source_location(self.obj)
        except Exception:  # noqa: BLE001
            return (
                f"Failed to inject {self.obj}. Missing arguments\n"
                f"Hint: Did you forget to make these dependencies connectable? "
                f"(Inherit from Connectable or "
                f"implement __connect__ and __disconnect__ methods)"
            )

        return (
            f"Failed to inject {self.obj}. \n"
            f"{object_location}\n\n"
            f"Missing arguments:\n"
            f"{missing_kwargs_str}\n\n"
            f"Hint: Did you forget to make these dependencies connectable? "
            f"(Inherit from Connectable or "
            f"implement __connect__ and __disconnect__ methods)"
        )


class InspectionError(InjectorError):
    def __init__(self, obj: Any):
        self.obj = obj

    def __str__(self) -> str:
        return self._build_error_message()

    def _build_error_message(self) -> str:
        object_location = _get_object_source_location(self.obj)

        return f"Failed to inspect {self.obj}. \n{object_location}\nSee the exception above"


def _get_object_source_location(obj: Callable[..., Any]) -> str:
    try:
        _, obj_line_number = inspect.getsourcelines(obj)
        source_file = inspect.getsourcefile(obj)
    except Exception:  # noqa: BLE001
        return ""

    return f"{source_file}:{obj_line_number}"



from __future__ import annotations

import asyncio
import threading
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from collections.abc import Coroutine

R = TypeVar("R")


@dataclass
class EventLoop:
    loop: asyncio.AbstractEventLoop
    running_outside: bool = False


class EventLoopGetter:
    def __init__(self) -> None:
        self._event_loop: asyncio.AbstractEventLoop | None = None
        self._lock = threading.Lock()
        self._loop_thread: threading.Thread | None = None

    def start(self) -> asyncio.AbstractEventLoop:
        with self._lock:
            if self._event_loop:
                return self._event_loop

            self._event_loop = asyncio.get_event_loop()
            self._loop_thread = threading.Thread(
                target=self._event_loop.run_forever,
                daemon=True,
            )
            self._loop_thread.start()
            return self._event_loop

    def __call__(self) -> asyncio.AbstractEventLoop:
        if self._event_loop:
            return self._event_loop

        return self.start()


def run_in_event_loop(
    coroutine: Coroutine[Any, Any, R],
    event_loop: EventLoop,
) -> R | None:
    # If we are inside event loop, we cannot obtain the result of an async task
    # because it will cause the event loop to hang.
    # In this case, we create a background task and return None.
    if event_loop.running_outside:
        event_loop.loop.create_task(coroutine)
        return None

    return asyncio.run_coroutine_threadsafe(coroutine, event_loop.loop).result()



from __future__ import annotations

from typing import Any
from unittest.mock import AsyncMock

from magic_di import ConnectableProtocol


def get_injectable_mock_cls(return_value: Any) -> type[ConnectableProtocol]:
    class ClientMetaclassMock(ConnectableProtocol):
        __annotations__ = {}

        def __new__(cls, *_: Any, **__: Any) -> Any:
            return return_value

    return ClientMetaclassMock


class InjectableMock(AsyncMock):
    """
    You can use this mock to override dependencies in tests
    and use AsyncMock instead of a real class instance

    Example:
        @pytest.fixture()
        def client():
          injector = DependencyInjector()

          with injector.override({Service: InjectableMock().mock_cls}):
            with TestClient(app) as client:
                yield client

        def test_http_handler(client):
          resp = client.post('/hello-world')

          assert resp.status_code == 200
    """

    @property
    def mock_cls(self) -> type[ConnectableProtocol]:
        return get_injectable_mock_cls(self)

    async def __connect__(self) -> None: ...

    async def __disconnect__(self) -> None: ...

    def __call__(self, *args: Any, **kwargs: Any) -> InjectableMock:
        return self.__class__(*args, **kwargs)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass
class Signature(Generic[T]):
    obj: T
    is_injectable: bool
    deps: dict[str, type] = field(default_factory=dict)
    kwargs: dict[str, type] = field(default_factory=dict)
    injector_arg: str | None = None



from __future__ import annotations

import asyncio
import os
import threading
from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

from celery import signals
from celery.loaders.app import AppLoader  # type: ignore[import-untyped]

from magic_di import DependencyInjector
from magic_di.celery._async_utils import EventLoop, EventLoopGetter, run_in_event_loop

if TYPE_CHECKING:
    from collections.abc import Callable

    from celery.loaders.base import BaseLoader


@runtime_checkable
class InjectedCeleryLoaderProtocol(Protocol):
    injector: DependencyInjector
    loaded: bool

    def on_worker_process_init(self) -> None: ...

    def get_event_loop(self) -> EventLoop: ...


def get_celery_loader(
    injector: DependencyInjector | None = None,
    event_loop_getter: Callable[[], asyncio.AbstractEventLoop] | None = None,
    log_fn: Callable[[str], None] = print,
) -> type[BaseLoader]:
    _injector = injector or DependencyInjector()
    _event_loop_getter = event_loop_getter or EventLoopGetter()

    class CeleryLoader(AppLoader):  # type: ignore[no-any-unimported, misc]
        def __init__(self, *args: Any, **kwargs: Any) -> None:
            super().__init__(*args, **kwargs)
            self.injector = _injector
            self.loaded = False
            self._event_loop: EventLoop | None = None
            self._lock = threading.Lock()
            self._event_loop_lock = threading.Lock()

        def on_worker_process_init(self) -> None:
            with self._lock:
                if self.loaded:
                    return

                log_fn(f"Trying to connect celery deps for pid {os.getpid()}...")
                run_in_event_loop(
                    self.injector.connect(),
                    self.get_event_loop(),
                )
                super().on_worker_process_init()
                log_fn(f"Celery deps are connected for pid {os.getpid()}")

                signals.worker_process_shutdown.connect(
                    self._disconnect,
                    weak=False,
                )
                self.loaded = True

        def get_event_loop(self) -> EventLoop:
            if self._event_loop is not None:
                return self._event_loop

            try:
                return EventLoop(asyncio.get_running_loop(), running_outside=True)
            except RuntimeError:
                # If a RuntimeError was raised,
                # it means that the task was called outside event loop.
                # And it also means that we can get the result of the coroutine.
                ...

            with self._event_loop_lock:
                if self._event_loop is None:
                    self._event_loop = EventLoop(
                        _event_loop_getter(),
                        running_outside=False,
                    )

                return self._event_loop

        def _disconnect(self, *_: Any, pid: str, **__: Any) -> None:
            log_fn(f"Trying to disconnect celery deps for pid {pid}...")
            run_in_event_loop(self.injector.disconnect(), self.get_event_loop())
            log_fn(f"Celery deps are disconnected for pid {pid}")

    return CeleryLoader



from __future__ import annotations

import inspect
from contextlib import asynccontextmanager
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Protocol,
    get_origin,
    runtime_checkable,
)

from fastapi.params import Depends

from magic_di._injector import DependencyInjector

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Callable, Iterator

    from fastapi import FastAPI, routing


@runtime_checkable
class RouterProtocol(Protocol):
    endpoint: Callable[..., Any]


@runtime_checkable
class FastAPIRouterProtocol(RouterProtocol, Protocol):
    dependencies: list[Depends]


def inject_app(
    app: FastAPI,
    *,
    injector: DependencyInjector | None = None,
    use_deprecated_events: bool = False,
) -> FastAPI:
    """
    Inject dependencies into a FastAPI application using the provided injector.

    This function sets up the FastAPI application to connect and disconnect the injector
    at startup and shutdown, respectively.

    This ensures that all dependencies are properly
    connected when the application starts, and properly disconnected when the application
    shuts down.

    If no injector is provided, a default injector is used.

    Args:
        app (FastAPI): The FastAPI application

        injector (DependencyInjector, optional): The injector to use for dependency injection.
                                                 If not provided, a default injector will be used.

        use_deprecated_events (bool, optional): Indicate whether the app should be injected
                                                with starlette events (which are deprecated)
                                                or use app lifespans.
                                                Important: Use this flag only
                                                if you still use app events,
                                                because if lifespans are defined,
                                                events will be ignored by Starlette.

    Returns:
        FastAPI: The FastAPI application with dependencies injected.
    """
    injector = injector or DependencyInjector()

    def collect_deps() -> None:
        _collect_dependencies(injector, app.router)

    app.state.dependency_injector = injector

    if use_deprecated_events:
        _inject_app_with_events(app, collect_deps_fn=collect_deps)
    else:
        _inject_app_with_lifespan(app, collect_deps_fn=collect_deps)

    return app


def _inject_app_with_lifespan(app: FastAPI, collect_deps_fn: Callable[[], None]) -> None:
    app_router: routing.APIRouter = app.router
    app_lifespan = app_router.lifespan_context

    @asynccontextmanager
    async def injector_lifespan(app: FastAPI) -> AsyncIterator[None]:
        collect_deps_fn()

        injector = app.state.dependency_injector
        await injector.connect()

        async with app_lifespan(app):
            yield

        await injector.disconnect()

    app_router.lifespan_context = injector_lifespan


def _inject_app_with_events(app: FastAPI, collect_deps_fn: Callable[[], None]) -> None:
    app.on_event("startup")(collect_deps_fn)
    app.on_event("startup")(app.state.dependency_injector.connect)
    app.on_event("shutdown")(app.state.dependency_injector.disconnect)


def _collect_dependencies(
    injector: DependencyInjector,
    app_router: routing.APIRouter,
) -> None:
    """
    It walks through all routers and collects all app dependencies to connect them later
    """
    for dependencies in app_router.dependencies:
        if not dependencies.dependency:
            continue

        for dependency in _find_fastapi_dependencies(dependencies.dependency):
            _inspect_and_lazy_inject(dependency, injector)

    for route in app_router.routes:
        if not isinstance(route, RouterProtocol):
            error_msg = (
                "Unexpected router class. "
                f"Router must contain .endpoint property with handler function: {route}"
            )
            raise TypeError(error_msg)

        if isinstance(route, FastAPIRouterProtocol):
            for dependencies in route.dependencies:
                for dependency in _find_fastapi_dependencies(dependencies.dependency):
                    _inspect_and_lazy_inject(dependency, injector)

        for dependency in _find_fastapi_dependencies(route.endpoint):
            _inspect_and_lazy_inject(dependency, injector)


def _inspect_and_lazy_inject(obj: object, injector: DependencyInjector) -> None:
    for dependency in injector.inspect(obj).deps.values():
        injector.lazy_inject(dependency)


def _find_fastapi_dependencies(dependency: Callable[..., Any]) -> Iterator[Callable[..., Any]]:
    """
    Recursively finds all FastAPI dependencies.
    It looks for FastAPI's Depends() in default arguments and in type annotations.
    """
    signature = inspect.signature(dependency)

    for param in signature.parameters.values():
        if isinstance(param.default, Depends) and param.default.dependency:
            yield from _find_fastapi_dependencies(param.default.dependency)

        type_hint = param.annotation
        if get_origin(type_hint) is not Annotated:
            continue

        for annotation in type_hint.__metadata__:
            if not isinstance(annotation, Depends):
                continue

            if annotation.dependency:
                yield from _find_fastapi_dependencies(annotation.dependency)

    yield dependency



from typing import Protocol, runtime_checkable


@runtime_checkable
class ConnectableProtocol(Protocol):
    """
    Interface for injectable clients.
    Adding these methods to your class will allow it to be dependency injectable.
    The dependency injector uses duck typing to check that the class
    implements the interface.
    This means that you do not need to inherit from this protocol.
    """

    async def __connect__(self) -> None: ...

    async def __disconnect__(self) -> None: ...


class Connectable:
    """
    You can inherit from this class to make the dependency visible for DI
    without adding these empty methods.
    """

    async def __connect__(self) -> None: ...

    async def __disconnect__(self) -> None: ...



from __future__ import annotations

import asyncio
import inspect
from typing import TYPE_CHECKING, TypeVar, overload

from magic_di import DependencyInjector

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable


T = TypeVar("T")


@overload
def inject_and_run(
    fn: Callable[..., Awaitable[T]],
    injector: DependencyInjector | None = None,
) -> T: ...


@overload
def inject_and_run(
    fn: Callable[..., T],
    injector: DependencyInjector | None = None,
) -> T: ...


def inject_and_run(
    fn: Callable[..., T],
    injector: DependencyInjector | None = None,
) -> T:
    """
    This function takes a callable, injects dependencies into it using the provided injector,
    and then runs the function. If the function is a coroutine, it will be awaited.

    The function itself is not asynchronous, but it uses asyncio.run
    to run the internal coroutine, so it is suitable for use in synchronous code.

    Args:
        fn (Callable): The function into which dependencies will be injected. This can be
                       a regular function or a coroutine function.
        injector (DependencyInjector, optional): The injector to use for dependency injection.
                                                 If not provided, a default injector will be used.

    Returns:
        Any: The return value of the function `fn` after dependency injection and execution.

    Raises:
        Any exceptions raised by the function `fn` or the injector will be propagated up to
        the caller of this function.
    """
    injector = injector or DependencyInjector()

    injected = injector.inject(fn)

    async def run() -> T:
        async with injector:
            if inspect.iscoroutinefunction(fn):
                return await injected()  # type: ignore[misc,no-any-return]

            return injected()

    return asyncio.run(run())



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Annotated, TypeVar

from fastapi import Depends, Request

from magic_di._injector import Injectable

if TYPE_CHECKING:
    from magic_di import DependencyInjector

T = TypeVar("T")


class FastAPIInjectionError(Exception): ...


class Provider:
    def __getitem__(self, obj: type[T]) -> type[T]:
        @lru_cache(maxsize=1)
        def get_dependency(injector: DependencyInjector) -> T:
            return injector.inject(obj)()

        def inject(request: Request) -> T:
            if not hasattr(request.app.state, "dependency_injector"):
                error_msg = (
                    "FastAPI application is not injected. Did you forget to add `inject_app(app)`?"
                )
                raise FastAPIInjectionError(error_msg)

            injector = request.app.state.dependency_injector
            return get_dependency(injector)

        return Annotated[obj, Depends(inject), Injectable]  # type: ignore[return-value]


if TYPE_CHECKING:
    from typing import Union as Provide  # hack for mypy # noqa: FIX004
else:
    Provide = Provider()
